import csv
import gettext
import json
import logging
import os
from collections import OrderedDict
from copy import deepcopy
from io import StringIO

from ocds_babel import TRANSLATABLE_SCHEMA_KEYWORDS, TRANSLATABLE_EXTENSION_METADATA_KEYWORDS  # noqa: E501
from ocds_babel.util import text_to_translate

try:
    from ocds_babel.translate_markdown import translate_markdown, translate_markdown_data  # noqa
except ImportError:
    pass

logger = logging.getLogger('ocds_babel')


def translate(configuration, localedir, language, headers, **kwargs):
    """
    Writes files, translating any translatable strings.

    For translated strings in schema files, replaces `{{lang}}` with the language code. Keyword arguments may specify
    additional replacements.
    """
    translators = {}

    for sources, target, domain in configuration:
        logger.info('Translating to {} using "{}" domain, into {}'.format(language, domain, target))

        if domain not in translators:
            translators[domain] = gettext.translation(
                domain, localedir, languages=[language], fallback=language == 'en')

        os.makedirs(target, exist_ok=True)

        for source in sources:
            basename = os.path.basename(source)
            with open(source) as r, open(os.path.join(target, basename), 'w') as w:
                if basename == 'extension.json':
                    method = translate_extension_metadata
                    kwargs.update(lang=language)
                elif source.endswith('.csv'):
                    method = translate_codelist
                    kwargs.update(headers=headers)
                elif source.endswith('.json'):
                    method = translate_schema
                    kwargs.update(lang=language)
                elif source.endswith('.md'):
                    method = translate_markdown
                else:
                    raise NotImplementedError(basename)
                w.write(method(r, translators[domain], **kwargs))


# This should roughly match the logic of `extract_codelist`.
def translate_codelist(io, translator, headers=[], **kwargs):
    """
    Accepts a CSV file as an IO object, and returns its translated contents in CSV format.
    """
    reader = csv.DictReader(io)

    fieldnames = [translator.gettext(fieldname) for fieldname in reader.fieldnames]
    rows = translate_codelist_data(reader, translator, headers, **kwargs)

    io = StringIO()
    writer = csv.DictWriter(io, fieldnames, lineterminator='\n')
    writer.writeheader()
    writer.writerows(rows)

    return io.getvalue()


def translate_codelist_data(source, translator, headers=[], **kwargs):
    """
    Accepts CSV rows as an iterable object (e.g. a list of dictionaries), and returns translated rows.
    """
    rows = []
    for row in source:
        data = {}
        for key, value in row.items():
            text = text_to_translate(value, key in headers)
            if text:
                value = translator.gettext(text)
            data[translator.gettext(key)] = value
        rows.append(data)
    return rows


# This should roughly match the logic of `extract_schema`.
def translate_schema(io, translator, **kwargs):
    """
    Accepts a JSON file as an IO object, and returns its translated contents in JSON format.
    """
    data = _json_load(io)

    data = translate_schema_data(data, translator, **kwargs)

    return _json_dumps(data)


def translate_schema_data(source, translator, **kwargs):
    """
    Accepts JSON data, and returns translated data.
    """
    def _translate_schema_data(data):
        if isinstance(data, list):
            for item in data:
                _translate_schema_data(item)
        elif isinstance(data, dict):
            for key, value in data.items():
                _translate_schema_data(value)
                text = text_to_translate(value, key in TRANSLATABLE_SCHEMA_KEYWORDS)
                if text:
                    data[key] = translator.gettext(text)
                    for old, new in kwargs.items():
                        data[key] = data[key].replace('{{' + old + '}}', new)

    data = deepcopy(source)
    _translate_schema_data(data)
    return data


# This should roughly match the logic of `extract_extension_metadata`.
def translate_extension_metadata(io, translator, lang='en', **kwargs):
    """
    Accepts an extension metadata file as an IO object, and returns its translated contents in JSON format.
    """
    data = _json_load(io)

    data = translate_extension_metadata_data(data, translator, lang, **kwargs)

    return _json_dumps(data)


def translate_extension_metadata_data(source, translator, lang='en', **kwargs):
    """
    Accepts extension metadata, and returns translated metadata.
    """
    data = deepcopy(source)

    for key in TRANSLATABLE_EXTENSION_METADATA_KEYWORDS:
        value = data.get(key)

        if isinstance(value, dict):
            value = value.get('en')

        text = text_to_translate(value)
        if text:
            data[key] = {lang: translator.gettext(text)}

    return data


def _json_load(io):
    return json.load(io, object_pairs_hook=OrderedDict)


def _json_dumps(data):
    return json.dumps(data, ensure_ascii=False, indent=2, separators=(',', ': '))



from docutils import core, nodes

skippable = (
    'visit_document',
    'depart_colspec',
    'depart_list_item',
    'depart_section',
    # Handled instead via `colspec` nodes.
    'depart_tgroup',
    'visit_tgroup',
    # Inline node
    'depart_emphasis',
    'depart_image',
    'depart_literal',
    'depart_pending_xref',
    'depart_raw',
    'depart_reference',
    'depart_strong',
    'visit_emphasis',
    'visit_image',
    'visit_literal',
    'visit_pending_xref',
    'visit_raw',
    'visit_reference',
    'visit_strong',
    # Text node
    'depart_Text',
    'visit_Text',
)


class MarkdownTranslator(nodes.NodeVisitor):
    def __init__(self, document, translator):
        nodes.NodeVisitor.__init__(self, document)

        # Whether we are writing to the output.
        self.writing = True

        # The writing context: 'block-quote', 'td' or 'th'.
        self.contexts = [None]
        # List item markers: '*' or '1.'.
        self.markers = []
        # Table column specifications.
        self.colspecs = []
        # For zebra tables.
        self.table_row_index = 1

        # The output.
        self.text = ''

        self.translator = translator

    @property
    def context(self):
        return self.contexts[-1]

    def append(self, text):
        if self.writing:
            self.text += text

    # See https://github.com/sphinx-doc/sphinx/blob/v2.2.1/sphinx/util/nodes.py#L252-L276
    def gettext(self, message):
        return self.translator.gettext(message.strip())

    def astext(self):
        return self.text

    # Otherwise, we need to implement a lot of empty methods to avoid exceptions.
    def __getattr__(self, name):
        def skip(*args):
            return

        def error(node):
            raise Exception(repr([name, {attr: getattr(node, attr) for attr in dir(node)}]))

        if name in skippable:
            return skip
        return error

    # System

    def depart_document(self, node):
        self.text = self.text[:-1]  # remove extra newline

    def visit_system_message(self, node):
        self.writing = False

    def depart_system_message(self, node):
        self.writing = True

    # Block

    def visit_block_quote(self, node):
        self.contexts.append('block-quote')

    def depart_block_quote(self, node):
        self.contexts.pop()

    def visit_paragraph(self, node):
        if self.context == 'block-quote':
            self.append('> ')

        message = self.gettext(node.rawsource)

        if self.context in ('td', 'th'):
            # Remove the opening "<p>" and closing "</p>\n".
            # See http://docutils.sourceforge.net/docs/api/publisher.html#publish-parts-details
            message = core.publish_parts(source=message, writer_name='html')['fragment'][3:-5]

        self.append(message)

    def depart_paragraph(self, node):
        if self.context not in ('td', 'th'):
            self.append('\n')
            if not self.markers:
                self.append('\n')

    def visit_literal_block(self, node):
        self.append('```{}\n'.format(node.attributes.get('language', '')))
        # Markdown code blocks (indented paragraphs) have no `rawsource`, but fenced code blocks do.
        if node.rawsource:
            text = node.rawsource
        else:
            text = node.astext()
        if not text.endswith('\n'):
            text += '\n'
        self.append(text)

    def depart_literal_block(self, node):
        self.append('```\n\n')

    def visit_section(self, node):
        self.append('#' * node.attributes['level'] + ' ')

    def visit_title(self, node):
        self.append(self.gettext(node.rawsource))

    def depart_title(self, node):
        self.append('\n\n')

    # Lists

    def visit_bullet_list(self, node):
        self.markers.append('*')

    def visit_enumerated_list(self, node):
        self.markers.append('1.')

    def depart_bullet_list(self, node):
        self.markers.pop()
        if not self.markers:
            self.append('\n')

    def depart_enumerated_list(self, node):
        self.markers.pop()
        if not self.markers:
            self.append('\n')

    def visit_list_item(self, node):
        self.append('  ' * (len(self.markers) - 1))
        self.append('{} '.format(self.markers[-1]))

    # Some parts copied from: docutils.writers._html_base, docutils.writers.html4css1, sphinx.writers.html

    # HTML

    def html_tag(self, node, tagname=None, suffix='\n', empty=None, **attributes):
        if node:
            parts = [node.tagname]
            atts = {k: v for k, v in node.attributes.items() if v}
        else:
            parts = [tagname]
            atts = {}

        atts.update(attributes)

        if empty:
            infix = ' /'
        else:
            infix = ''

        for name, value in atts.items():
            parts.append('{}="{}"'.format(name.lower(), value))

        self.append('<{}{}>'.format(' '.join(parts), infix) + suffix)

    def close_html_tag(self, node, tagname=None, suffix='\n'):
        self.append('</{}>'.format(tagname or node.tagname) + suffix)

    def write_colspecs(self):
        if self.colspecs:
            self.html_tag(None, 'colgroup')
            width = sum(node['colwidth'] for node in self.colspecs)
            for node in self.colspecs:
                colwidth = int(node['colwidth'] * 100.0 / width + 0.5)
                self.html_tag(None, 'col', empty=True, width='{}%'.format(colwidth))
            self.colspecs = []
            self.close_html_tag(None, 'colgroup')

    # Tables

    def visit_table(self, node):
        self.html_tag(node, border='1', CLASS='docutils')
        self.table_row_index = 1

    def depart_table(self, node):
        self.close_html_tag(node, suffix='\n\n')

    def visit_colspec(self, node):
        self.colspecs.append(node)

    def visit_thead(self, node):
        self.write_colspecs()
        self.html_tag(node, valign='bottom')

    def depart_thead(self, node):
        self.close_html_tag(node)

    def visit_tbody(self, node):
        self.write_colspecs()
        self.html_tag(node, valign='top')

    def depart_tbody(self, node):
        self.close_html_tag(node)

    def visit_row(self, node):
        self.html_tag(None, 'tr', CLASS='row-odd' if self.table_row_index % 2 else 'row-even')
        self.table_row_index += 1

    def depart_row(self, node):
        self.close_html_tag(None, 'tr')

    def visit_entry(self, node):
        if isinstance(node.parent.parent, nodes.thead):
            tagname = 'th'
            atts = {'class': 'head'}
        else:
            tagname = 'td'
            atts = {}
        self.contexts.append(tagname)  # needs to be `tagname`, for `depart_entry` to use
        self.html_tag(None, tagname, suffix='', **atts)

    def depart_entry(self, node):
        self.close_html_tag(None, self.contexts.pop())



from docutils.nodes import literal_block
from docutils.parsers.rst import Directive


class NullDirective(Directive):
    has_content = True

    def run(self):
        return [literal_block(self.block_text, language='eval_rst')]



import logging

from docutils.frontend import OptionParser
from docutils.io import InputError
from docutils.parsers.rst import Parser, directives
from docutils.utils import new_document, SystemMessage
from recommonmark.parser import CommonMarkParser
from recommonmark.transform import AutoStructify
from sphinx.application import Sphinx
from sphinx.util.docutils import docutils_namespace, patch_docutils

from ocds_babel.directives import NullDirective
from ocds_babel.markdown_translator import MarkdownTranslator

logger = logging.getLogger('ocds_babel')


def translate_markdown(io, translator, **kwargs):
    """
    Accepts a Markdown file as an IO object, and returns its translated contents in Markdown format.
    """
    name = io.name
    text = io.read()

    return translate_markdown_data(name, text, translator, **kwargs)


def translate_markdown_data(name, text, translator, **kwargs):
    """
    Accepts a Markdown file as its filename and contents, and returns its translated contents in Markdown format.
    """
    # This only needs to be run once, but is inexpensive.
    for directive_name in ('csv-table-no-translate', 'extensiontable'):
        directives.register_directive(directive_name, NullDirective)

    with patch_docutils(), docutils_namespace():
        # sphinx-build -b html -q -E …
        app = Sphinx('.', None, 'outdir', '.', 'html', status=None, freshenv=True)
        # Avoid "recommonmark_config not setted, proceed default setting".
        app.add_config_value('recommonmark_config', {}, True)

        # From code comment in `new_document`.
        settings = OptionParser(components=(Parser,)).get_default_values()
        # Get minimal settings for `AutoStructify` to be applied.
        settings.env = app.builder.env

        document = new_document(name, settings)
        CommonMarkParser().parse(text, document)

        # To translate messages in `.. list-table`.
        try:
            AutoStructify(document).apply()
        except SystemMessage as e:
            context = e.__context__
            if isinstance(context, InputError) and context.strerror == 'No such file or directory':  # csv-table
                logger.warning(e)

        visitor = MarkdownTranslator(document, translator)
        document.walkabout(visitor)

        return visitor.astext()
